// kernel call graph profiler - watches system calls and follows paths taken

#include <iostream>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <limits.h>
#include <wait.h>

#include <vector>
#include <ext/hash_set>
#include <ext/hash_map>
#include <utility>
using namespace std;
using namespace __gnu_cxx;

#include "kapi.h"
kapi_manager kmgr;

//template class kapi_vector< pair<kptr_t, kptr_t> >;

void error_exit(int retval)
{
    kmgr.detach();
    exit(retval);
}

time_t callee_watching_delay = 15;
time_t visualization_delay = 120;
time_t profile_delay = 5;

// number of levels deep to pre-instrument callees of "hot" functions
unsigned max_deep_start_level = 2;

unsigned the_app_pid = 0;

typedef uint32_t uptr_t;

uptr_t ctr_cache_base_addr = 0;
unsigned next_avail_ctr = 0;
const unsigned max_num_ctrs = 2048;
unsigned ctr_cache_size = 8192; //max_num_ctrs * sizeof(unsigned);
char ctr_cache_user_copy[8192];

void createCtrCache()
{
    // allocate it
    ctr_cache_base_addr = kmgr.malloc(ctr_cache_size);
    if (ctr_cache_base_addr == 0) {
        fprintf(stderr, "kmgr.malloc error\n");
        error_exit(-1);
    }

    // zero it
    memset(ctr_cache_user_copy, 0, ctr_cache_size);
    if (kmgr.memcopy(ctr_cache_user_copy, ctr_cache_base_addr, 
                      ctr_cache_size) < 0) {
        fprintf(stderr, "kmgr.memcopy error initializing ctr_cache\n");
        error_exit(-1);
    }
}

void dumpCtrCache()
{
    unsigned *curr = (unsigned*)&ctr_cache_user_copy[0];
    unsigned curr_ndx = 0;
    printf("Contents of ctr cache user copy:\n");
    while(curr_ndx < next_avail_ctr) {
       printf("%d ", *curr++);       
       if(++curr_ndx % 16 == 0)
          printf("\n");
    }
    printf("\n");
}

void freeCtrCache()
{
    memset(ctr_cache_user_copy, 0, ctr_cache_size);
    if( kmgr.free(ctr_cache_base_addr) < 0)
        fprintf(stderr, "kmgr.free error\n");
}

void readCtrs()
{  
    if(next_avail_ctr) { 
        if (kmgr.memcopy(ctr_cache_base_addr, ctr_cache_user_copy, 
                         next_avail_ctr*sizeof(unsigned)) < 0) {
            fprintf(stderr, "kmgr.memcopy error reading ctr_cache\n");
            error_exit(-1);
        }
    }
}

uptr_t ndx2CtrAddr(unsigned ndx)
{
    assert(ctr_cache_base_addr != 0);
    assert(ndx != (unsigned)-1);
    return ctr_cache_base_addr + ndx*sizeof(unsigned);
}

unsigned nextCtrNdx()
{
    unsigned next_ctr = (unsigned)-1;
    if(next_avail_ctr < max_num_ctrs) {
        next_ctr = next_avail_ctr++;
    }
    else
        fprintf(stderr, "nextCtr() - used up ctr cache\n");  
    return next_ctr;
}

unsigned getCtrValue(unsigned ctr_ndx)
{
    if(ctr_ndx == (unsigned)-1)
        return 0;

    return *(unsigned*)&ctr_cache_user_copy[ctr_ndx*sizeof(unsigned)];
}

kapi_snippet generateCtrInc(unsigned ctr_ndx)
{
    assert(the_app_pid != 0);
    kptr_t ctr_addr = ndx2CtrAddr(ctr_ndx);
    kapi_int_variable theCtr(ctr_addr);
    kapi_arith_expr inc_ctr(kapi_atomic_assign, theCtr, 
                            kapi_arith_expr(kapi_plus, theCtr, 
                                            kapi_const_expr(1)));
    kapi_bool_expr pid_test(kapi_eq, kapi_const_expr(the_app_pid), 
                            kapi_pid_expr()); 
    return kapi_if_expr(pid_test, inc_ctr);
}

void log_string(const char *s)
{
    FILE *f = fopen("instlist.log", "at");
    assert(f);
    fprintf(f, "%s\n", s);
    fclose(f);
}

unsigned getNextId()
{
    static unsigned global_node_id = 0;
    return global_node_id++;
}

typedef struct {
    uptr_t caller_addr;
    uptr_t callee_addr;
    uptr_t site_addr;
    unsigned ctr_ndx;
    unsigned inst_id;
    bool is_indirect; 
} edge_t;

struct eqedge {
    bool operator()(const edge_t &e1, const edge_t &e2) const {
	return (e1.site_addr == e2.site_addr &&
		e1.callee_addr == e2.callee_addr);
    }
};

struct hash_edge {
    size_t operator()(const edge_t &e) const {
	return (e.caller_addr + e.callee_addr);
    }
};

struct hash_uptr_t {
    size_t operator()(const uptr_t &u) const {
	return u;
    }
};

class node_t;
typedef hash_map<uptr_t, node_t *, hash_uptr_t> node_map_t;
typedef hash_set<edge_t, hash_edge, eqedge> edge_set_t;


class node_t {
    unsigned id; // sequential id, mostly for drawing
    bool processed;
    bool hot;
    bool instrumentable;
    bool callees_done;
    const char *name;
 public:
    node_t(unsigned iid, bool iable, const char *iname) :
	id(iid), processed(false), hot(false), instrumentable(iable),
	callees_done(false), name(strdup(iname)) {
    }
    void setProcessed() {
	processed = true;
    }
    unsigned getId() const {
	return id;
    }
    bool isProcessed() const {
	return processed;
    }
    bool isHot() const {
	return hot;
    }
    void setHot() {
	hot = true;
    }
    bool isInstrumentable() const {
	return instrumentable;
    }
    void setCalleesDone() {
        callees_done = true;
    }
    bool calleesDone() {
        return callees_done;
    }
   
    const char *getName() const {
	return name;
    }

    virtual unsigned getEntryCtrValue() const = 0;
    virtual int unInstrument() = 0;
    virtual int doCallees(node_map_t *nodes, edge_set_t *edges, 
                          vector< pair<kptr_t, kptr_t> > &indirects,
                          unsigned deep_start_level) = 0;
    virtual const char *getColor() const = 0;
};

class knode_t : public node_t {
    kapi_function kfunc;
    unsigned ctr_ndx;
    int entry_id;
    int getMyCallees(vector<edge_t> *edges, 
                     vector< pair<kptr_t, kptr_t> > &indirects);
public:
    knode_t(unsigned iid, const char *iname,
	    const kapi_function &ifunc, unsigned ictr_ndx, int ientry_id) :
	node_t(iid, !ifunc.isUnparsed(), iname), kfunc(ifunc),
        ctr_ndx(ictr_ndx), entry_id(ientry_id) {
    }

    unsigned getEntryCtrValue() const { return getCtrValue(ctr_ndx); }
    int unInstrument();
    int doCallees(node_map_t *nodes, edge_set_t *edges, 
                  vector< pair<kptr_t, kptr_t> > &indirects,
                  unsigned deep_start_level);
    const char *getColor() const;
    const kapi_function &getFunc() const { return kfunc; }
};

int knode_t::getMyCallees(vector<edge_t> *edges, 
                          vector< pair<kptr_t, kptr_t> > &indirects)
{
   //    printf("knode_t::getMyCallees()\n");
    kapi_vector<kptr_t> regCallsFromAddr, regCallsToAddr;
    kapi_vector<kptr_t> interprocFromAddr, interprocToAddr;
    kapi_vector<kptr_t> unanalyzableCallsFromAddr;
    
    if (kfunc.getCallees(NULL, // all blocks
                         &regCallsFromAddr, &regCallsToAddr,
                         &interprocFromAddr, &interprocToAddr, 
                         &unanalyzableCallsFromAddr) < 0) {
	fprintf(stderr, "kfunc.getCallees error\n");
	return (-1);
    }
    const kptr_t fnEntryAddr = kfunc.getEntryAddr();
    const unsigned numRegCalls = regCallsFromAddr.size();
    for (unsigned i=0; i<numRegCalls; i++) {
	edge_t edge;
	edge.caller_addr = fnEntryAddr;
	edge.callee_addr = regCallsToAddr[i];
	edge.site_addr = regCallsFromAddr[i];
        edge.ctr_ndx = nextCtrNdx();
        edge.inst_id = (unsigned)-1;
        edge.is_indirect = false;
	edges->push_back(edge);
    }

    const unsigned numIndirectCalls = unanalyzableCallsFromAddr.size();
    for (unsigned i=0; i<numIndirectCalls; i++) {
       printf("adding indirect call (%08lx, %08lx)\n", 
              unanalyzableCallsFromAddr[i], fnEntryAddr);
        indirects.push_back(
                   make_pair<kptr_t, kptr_t>(unanalyzableCallsFromAddr[i], 
                                             fnEntryAddr));
    }
    return 0;
}

node_t* addNodeIfNew(const uptr_t fn_addr, node_map_t *nodes, 
                     bool do_edge_counts)
{
   //    printf("addNodeIfNew()\n");
    const unsigned maxlen = 255;
    char tmpName[maxlen];

    if (nodes->find(fn_addr) == nodes->end()) {
        kapi_function kcallee;
        if (kmgr.findFunctionByAddress(fn_addr, &kcallee) < 0) {
            fprintf(stderr, "Could not find function at 0x%x\n", fn_addr);
            return NULL;
        }
        if (kcallee.getName(tmpName, maxlen) < 0) {
            fprintf(stderr, "Name is too long\n");
            return NULL;
        }

        unsigned ctr_ndx = (unsigned)-1;
        int entry_id = -1;
	    
        if (kcallee.isUnparsed()) {
            fprintf(stderr, "Not instrumenting %s\n", tmpName);
        }
        else if(!do_edge_counts) {
            ctr_ndx = nextCtrNdx();
            kapi_vector<kapi_point> entries;
            if (kcallee.findEntryPoint(&entries) < 0) {
                fprintf(stderr, "kcallee.findEntryPoint error\n");
                return NULL;
            }
            assert(entries.size() == 1);

            if(ctr_ndx != (unsigned)-1) {
                kapi_snippet inc_ctr_expr = generateCtrInc(ctr_ndx);
                if ((entry_id = kmgr.insertSnippet(inc_ctr_expr,
                                                    entries[0])) < 0) {
                    fprintf(stderr, "kmgr.insertSnippet(start) error\n");
                    return NULL;
                }
            }
        }
        printf("addNodeIfNew: adding node (%s, %08lx, ctr_ndx=%d)\n", 
               tmpName, fn_addr, ctr_ndx);
        node_t *node = new knode_t(getNextId(), tmpName,
                                   kcallee, ctr_ndx, entry_id);
        if(do_edge_counts) {
           node->setProcessed();
           node->setHot();
        }
        nodes->insert(pair<uptr_t,node_t*>(fn_addr, node));
        return node;
    }
    return NULL;
}

int knode_t::doCallees(node_map_t *nodes, edge_set_t *edges,
                       vector< pair<kptr_t, kptr_t> > &indirects,
                       unsigned deep_start_level)
{
    vector<edge_t> candidates;
    //printf("knode_t::doCallees()\n");

    if(! calleesDone()) {

        if (getMyCallees(&candidates, indirects) < 0) {
	    return (-1);
        }

        vector<edge_t>::const_iterator eiter = candidates.begin();
        edge_set_t::const_iterator edges_end = edges->end();
        for (; eiter != candidates.end(); eiter++) {
	    // Add the edge if it is not there yet
            edge_set_t::const_iterator match = edges->find(*eiter);
            if (match == edges_end) {
                printf("doCallees: adding edge (%08lx, %08lx, %08lx)\n",
                       eiter->caller_addr, eiter->site_addr, eiter->callee_addr);
                edges->insert(*eiter);
            }
            else {
                printf("doCallees: edge (%08lx, %08lx, %08lx) matches (%08lx, %08lx, %08lx), not adding\n", 
                       eiter->caller_addr, eiter->site_addr, eiter->callee_addr, 
                       match->caller_addr, match->site_addr, match->callee_addr);
               
            }
            // Start profiling callee, if not already identified
            node_t *new_node = addNodeIfNew(eiter->callee_addr, nodes, false);

            // Deep start - pre-instrument down static paths
            if(new_node != NULL) {
                if(deep_start_level < max_deep_start_level) {
                    new_node->doCallees(nodes, edges, indirects, 
                                        deep_start_level+1);
                }
            }
        }
    }
    setCalleesDone();
    return 0;
}

int knode_t::unInstrument()
{
    if (entry_id == -1) {
	// Some functions are not instrumentable. Do not try to
	// un-instrument them.
	return 0;
    }

    //char str[255];
    //sprintf(str, "uninstrumenting %s...", getName());
    //log_string(str);

    if (kmgr.removeSnippet(entry_id) < 0) {
	fprintf(stderr, "removeSnippet error\n");
	return (-1);
    }
    entry_id = -1; //prevent another attempt to uninstrument

    //log_string("done");
    return 0;
}

const char *knode_t::getColor() const
{
    return "white";
}

int drawGraphIfTime(const node_map_t &nodes, const edge_set_t &edges);

bool doProfile(node_map_t *nodes, edge_set_t *edges, 
               vector< pair<kptr_t, kptr_t> > &indirects)
{
    bool is_done = true;
    //printf("doProfile()\n");

    // update local copy of ctr cache
    readCtrs();

    drawGraphIfTime(*nodes, *edges);

    node_map_t::iterator iter = nodes->begin();
    for (; iter != nodes->end(); iter++) {
	node_t *node = iter->second;
	if (!node->isProcessed()) {
	    is_done = false;
            unsigned val = node->getEntryCtrValue();
	    if (val >= 1) {
		node->setProcessed();
                node->setHot();
		if (node->unInstrument() < 0) {
		    return true;
                }
                node->doCallees(nodes, edges, indirects, 1);
	    }
	}
    }
    return is_done;
}

unsigned num_watched = 0;

void watchIndirects(vector< pair<kptr_t, kptr_t> > &callsites,
	            time_t &start_time)
{
    int rv;
    num_watched = callsites.size();
    //printf("watchIndirects - num_watched = %d\n", num_watched);
    for (unsigned i=0; i<num_watched; i++) {
	const kptr_t siteAddr = callsites[i].first;
        //printf("- watching %08lx\n", siteAddr);
	if ((rv = kmgr.watchCalleesAt(siteAddr)) < 0) {
	    fprintf(stderr, "kmgr.watchCalleesAt error %d for site 0x%08x\n", rv, siteAddr);
	    return;
	}
    }
    if(num_watched)
        start_time = time(NULL);
}

void unwatchIndirects(vector< pair<kptr_t, kptr_t> > &callsites, 
                      node_map_t *nodes, edge_set_t *edges)
{
   //printf("unwatchIndirects - num_watched = %d\n", num_watched);
    for (unsigned i=0; i<num_watched; i++) {
	const kptr_t siteAddr = callsites[i].first;
        const kptr_t callerAddr = callsites[i].second;
        //printf("- unwatching %08lx in %08lx\n", siteAddr, callerAddr);
	if (kmgr.unWatchCalleesAt(siteAddr) < 0) {
	    fprintf(stderr, "kmgr.unWatchCalleesAt error\n");
	    return;
	}
	kapi_vector<kptr_t> callees;
	kapi_vector<unsigned> counts;
	if (kmgr.getCalleesForSite(siteAddr, &callees, &counts) < 0) {
	    fprintf(stderr, "kmgr.getCalleesForSite error\n");
	    return;
	}
        //printf(" site has %d callees\n", callees.size());
	kapi_vector<kptr_t>::const_iterator iter = callees.begin();
        edge_set_t::const_iterator edges_end = edges->end();
	for (; iter != callees.end(); iter++) {
            edge_t edge;
            edge.caller_addr = callerAddr;
            //printf("unwatchIndirects: *iter/callee is %08lx\n", *iter);
            edge.callee_addr = *iter;
            edge.site_addr = siteAddr;
            edge_set_t::const_iterator match = edges->find(edge);
            if (match == edges_end) {
                edge.inst_id = (unsigned)-1;
                edge.is_indirect = true;
                edge.ctr_ndx = nextCtrNdx();
                printf("unwatchIndirects: adding edge (%08lx, %08lx, %08lx)\n",
                       edge.caller_addr, edge.site_addr, edge.callee_addr);
                edges->insert(edge);
            }
            // else {
//                printf("unwatchIndirects: edge (%08lx, %08lx, %08lx) matches (%08lx, %08lx, %08lx), not adding\n", 
//                       edge.caller_addr, edge.site_addr, edge.callee_addr, 
//                       match->caller_addr, match->site_addr, match->callee_addr); 
//             }
            // start profiling callee, if not already identified
            addNodeIfNew(*iter, nodes, false);
	}
    }
}

void createRootNode(const char *root_mod, const char *root_func,
                   node_map_t &nodes, 
                   vector< pair<kptr_t, kptr_t> > &indirect_calls)
{
    int rv;

    printf("Locating the root function '%s/%s'\n", root_mod, root_func);
    // Find the module of interest, initialize kmod
    kapi_module kmod;
    if ((rv = kmgr.findModule(root_mod, &kmod)) < 0) {
        cerr << "findModule() error\n";
        error_exit(rv);
    }

    // Find the function of interest, initialize kfunc
    kapi_function kfunc;
    if ((rv = kmod.findFunction(root_func, &kfunc)) < 0) {
        cerr << "findFunction() error\n";
        error_exit(rv);
    }

    printf("Creating root callgraph node\n");
    const kptr_t fn_addr = kfunc.getEntryAddr();
    unsigned ctr_ndx = nextCtrNdx();
    node_t *node = new knode_t(getNextId(), root_func,
                               kfunc, ctr_ndx, -1);
    node->setHot();
    node->setProcessed();
    nodes.insert(pair<uptr_t,node_t*>(fn_addr, node));

    kapi_vector<kptr_t> regCallsFromAddr, regCallsToAddr;
    kapi_vector<kptr_t> interprocFromAddr, interprocToAddr;
    kapi_vector<kptr_t> unanalyzableCallsFromAddr;
    
    if ((rv = kfunc.getCallees(NULL,
                               &regCallsFromAddr, &regCallsToAddr,
                               &interprocFromAddr, &interprocToAddr, 
                               &unanalyzableCallsFromAddr)) < 0) {
        cerr << "getCallees error\n";
        error_exit(rv);
    }
    const unsigned numIndirectCalls = unanalyzableCallsFromAddr.size();
    for (unsigned i=0; i<numIndirectCalls; i++) {
        printf("adding indirect call (%08lx, %08lx)\n", 
               unanalyzableCallsFromAddr[i], fn_addr);
        indirect_calls.push_back(
                   make_pair<kptr_t, kptr_t>(unanalyzableCallsFromAddr[i], 
                                             fn_addr));
    }
}

vector< pair<int, uptr_t> > prev_node_ids;

uptr_t findPrevNodeAddr(int id)
{
   vector< pair<int, uptr_t> >::const_iterator iter = prev_node_ids.begin();
   vector< pair<int, uptr_t> >::const_iterator end = prev_node_ids.end();
   for(; iter != end; iter++) {
      if((*iter).first == id)
         return (*iter).second;
   }
   return 0;
}

void initializeFromPrevRun(const char *file, 
                           node_map_t &nodes,
                           edge_set_t &edges,
                           vector< pair<kptr_t, kptr_t> > &indirect_calls,
                           bool do_edge_counts)
{
    int i;
    vector<kptr_t> indirects_added;

    assert(file != NULL);
    printf("Reading previous run result file and initializing callgraph\n");
    FILE *f = fopen(file, "r");
    if (f == 0) {
        fprintf(stderr, "fopen error\n");
        error_exit(-1);
    }
    char line_buffer[256];
    // skip first 4 lines of dot directives
    for(i = 0; i < 4; i++) {
        if(NULL == fgets(line_buffer, 256, f)) {
            cerr << "error on file read\n";
            fclose(f);
            error_exit(-1);
        }
    }
    // read nodes, then edges
    while(fgets(line_buffer, 256, f) != NULL) {
        int id1, id2;
        unsigned count;
        uptr_t addr;
        char name[80];
        char shape[10];
        char color[15];
        edge_t edge;
        if(sscanf(line_buffer,
 "a%d [ label=\"%s %x\" ,shape=%s ,style=filled ,fillcolor=%s ,color=black ];",
                  &id1, name, &addr, shape, color) == 5) {
            addNodeIfNew(addr, &nodes, do_edge_counts);
            prev_node_ids.push_back(make_pair<int, uptr_t>(id1, addr));
        }
        else if(sscanf(line_buffer, 
                       "a%d->a%d [ label=\"%x %d\" ,style=%s ];", 
                       &id1, &id2, &addr, &count, shape) == 5) {
            edge.caller_addr = findPrevNodeAddr(id1);
            edge.callee_addr = findPrevNodeAddr(id2);
            edge.site_addr = addr;
            edge.ctr_ndx = nextCtrNdx();
            edge.inst_id = (unsigned)-1;
            edge.is_indirect = true;
            vector<kptr_t>::iterator enditer = indirects_added.end();
            if( find(indirects_added.begin(), enditer, addr) == enditer) {
                indirects_added.push_back(addr);
                indirect_calls.push_back(
                        make_pair<kptr_t, kptr_t>(addr, edge.caller_addr));
            }
            printf("adding initial indirect edge (%08lx, %08lx, %08lx, ctr_ndx=%d)\n",
                   edge.caller_addr, edge.site_addr, edge.callee_addr, edge.ctr_ndx);
            edges.insert(edge);
        }
        else if(sscanf(line_buffer, 
                       "a%d->a%d [ label=\"%x %d\" ];", 
                       &id1, &id2, &addr, &count) == 4) {
            edge.caller_addr = findPrevNodeAddr(id1);
            edge.callee_addr = findPrevNodeAddr(id2);
            edge.site_addr = addr;
            edge.ctr_ndx = nextCtrNdx();
            edge.inst_id = (unsigned)-1;
            edge.is_indirect = false;
            printf("adding initial direct edge (%08lx, %08lx, %08lx, ctr_ndx=%d)\n",
                   edge.caller_addr, edge.site_addr, edge.callee_addr, edge.ctr_ndx);
            edges.insert(edge);
        }
        else {
            if(strcmp(line_buffer, "}\n") != 0)
                printf("WARNING - ignoring non-spec line: %s", line_buffer);
        }
    }


    if(do_edge_counts) { // instrument edges now
        printf("Instrumenting edges to collect execution counts\n");
        edge_set_t::iterator eiter = edges.begin();
        for (; eiter != edges.end(); eiter++) {
            node_map_t::const_iterator caller = nodes.find(eiter->caller_addr);
            node_map_t::const_iterator callee = nodes.find(eiter->callee_addr);
            if (caller == nodes.end() || callee == nodes.end()) {
                fprintf(stderr, "Function not found, ignoring the edge\n");
                continue;
            }
            const knode_t *rnode = (knode_t*)caller->second;
            const knode_t *enode = (knode_t*)callee->second;
            int inst_id;
            if(eiter->ctr_ndx != (unsigned)-1) { 
                
                if(eiter->is_indirect) {
                    /* in order to get an edge count for indirect callsites
                       we must instrument the callee entry point and check the
                       function return address on the stack to see if it
                       matches the instruction following the callsite. */
                   
                    kapi_point entry;
                    kmgr.createInstPointAtAddr(eiter->callee_addr, &entry);

                    kapi_disass_object disass_call;
                    kmgr.getDisassObject(eiter->site_addr, eiter->site_addr+10,
                                         &disass_call);
                    unsigned call_insn_len = ((disass_call.begin())->begin())->getNumBytes();

                    kapi_int_variable ret_addr(kapi_raw_register_expr(20/*%esp*/));
                    kapi_bool_expr check_return(kapi_eq, ret_addr, 
                                                kapi_const_expr(eiter->site_addr + call_insn_len));
                    kapi_if_expr inc_ctr_expr(check_return,  
                                              generateCtrInc(eiter->ctr_ndx));
     
                    printf("- indirect edge 0x%08lx -> 0x%08lx\n", 
                           eiter->site_addr, eiter->callee_addr);
                    if ((inst_id = kmgr.insertSnippet(inc_ctr_expr,
                                                      entry)) < 0) {
                        fprintf(stderr, "insertSnippet(indir-edge) error\n");
                        error_exit(inst_id);
                    }
                }
                else { // just instrument the callsite

                    kapi_point callsite;
                    kmgr.createInstPointAtAddr(eiter->site_addr, &callsite);

                    kapi_snippet inc_ctr_expr = generateCtrInc(eiter->ctr_ndx);
                    printf("- direct edge 0x%08lx -> 0x%08lx\n", 
                           eiter->site_addr, eiter->callee_addr);
                    if ((inst_id = kmgr.insertSnippet(inc_ctr_expr,
                                                      callsite)) < 0) {
                        fprintf(stderr, "insertSnippet(dir-edge) error\n");
                        error_exit(inst_id);
                    }
                }
                edge_t &the_edge = const_cast<edge_t&>(*eiter); 
                the_edge.inst_id = inst_id;

            }
	}
    }
}

char *cg_file_name = NULL;

int dumpCallGraph(const node_map_t &nodes, const edge_set_t &edges, 
                  bool do_edge_counts)
{
    //const bool doCold = true;
    const bool doCold = false;
    //printf("dumpCallGraph()\n");

    assert(cg_file_name != NULL);
    FILE *f = fopen(cg_file_name, "wt");
    if (f == 0) {
	fprintf(stderr, "fopen error\n");
	return (-1);
    }
    fprintf(f,"digraph callgraph {\ncenter=true;\n"
	    "size=\"8,11.5\"\nratio=fill;\n");
// Append "rotate=90;\n" for landscape printing
    node_map_t::const_iterator niter = nodes.begin();
    for (; niter != nodes.end(); niter++) {
	node_t *node = niter->second;
	unsigned funcId = node->getId();
	bool hot =  node->isHot();
	bool instrumentable = node->isInstrumentable();
	const char *funcName = node->getName();
	const char *color = node->getColor();

        char name_addr[80];
        sprintf(name_addr, "%s %08x", funcName, niter->first);
	if (instrumentable && hot) {
	    fprintf(f, "a%u [ label=\"%s\" ,shape=ellipse ,"
		    "style=filled ,fillcolor=palegreen ,color=black ];\n",
		    funcId, name_addr);
	}
	else if (!instrumentable && doCold) {
	    fprintf(f, "a%u [ label=\"%s\" ,shape=octagon ,"
                    "style=filled ,fillcolor=gray ,color=black ];\n",
                    funcId, name_addr);
	}
	else if (!hot && doCold) {
	    fprintf(f, "a%u [ label=\"%s\" ,shape=box ,"
                    "style=filled ,fillcolor=lightblue ,color=black ];\n",
                    funcId, name_addr);
	}
    }

    edge_set_t::const_iterator eiter = edges.begin();
    for (; eiter != edges.end(); eiter++) {
	node_map_t::const_iterator caller = nodes.find(eiter->caller_addr);
	node_map_t::const_iterator callee = nodes.find(eiter->callee_addr);
	if (caller == nodes.end() || callee == nodes.end()) {
	    fprintf(stderr, "Function not found, ignoring the edge\n");
	    continue;
	}
	const node_t *rnode = caller->second;
	const node_t *enode = callee->second;
	if (doCold || (rnode->isInstrumentable() && rnode->isHot() &&
		       enode->isInstrumentable() && enode->isHot())) {
	    if(eiter->is_indirect)
                fprintf(f,"a%u->a%u [ label=\"%08x %d\" ,style=dashed ];\n", 
                        rnode->getId(), enode->getId(), eiter->site_addr,
                        (do_edge_counts ? getCtrValue(eiter->ctr_ndx) : 0));
            else
	        fprintf(f,"a%u->a%u [ label=\"%08x %d\" ];\n", 
                        rnode->getId(), enode->getId(), eiter->site_addr,
                        (do_edge_counts ? getCtrValue(eiter->ctr_ndx) : 0));
	}
    }

    fprintf(f,"}\n");
    fclose(f);

//    if (system("dot -Tps graph.dt -o graph1.ps") < 0) {
//	perror("system error");
//	return (-1);
//    }

//    if (system("cp graph1.ps graph.ps") < 0) {
//	perror("system error");
//	return (-1);
 //   }

    return 0;
}

int drawGraphIfTime(const node_map_t &nodes, const edge_set_t &edges)
{
    static time_t last_time = 0;
    time_t t = time(NULL);
    if (t - last_time > visualization_delay) {
	if (dumpCallGraph(nodes, edges, false) < 0) {
	    return (-1);
	}
	last_time = t;
    }
    return 0;
}

const char *base_name = NULL;
int usage()
{
    assert(base_name != NULL);
    cerr << "Usage: " << base_name << " <port#> <pid> { --start <module_name> <func_name> | --continue prev_run_results_file | --count-edges prev_run_results_file ] }\n";
    return (-1);
}

int main(int argc, char **argv, char **envp)
{
    pid_t pid;
    int i, rv = 0;
    time_t watch_start = 0;

    base_name = argv[0];

    if (argc < 5 || argc > 6)
	return usage();

    const char *host = "localhost";
    int port = atoi(argv[1]);
    pid = atoi(argv[2]);
    const char *run_type = argv[3];

    const char *start_mod = NULL;
    const char *start_func = NULL;
    const char *prev_run_file = NULL;
    bool do_edge_counts = false;
    bool start_run = true;

    if (! strcmp(run_type, "--start")) {
        if (argc == 6) {
            start_mod = argv[4];
            start_func = argv[5];
        }
        else
            return usage();
    }
    else if (! strcmp(run_type, "--continue")) {
        start_run = false;
        if(argc == 5)
           prev_run_file = argv[4];
        else
           return usage();
    }
    else if (! strcmp(run_type, "--count-edges")) {
        start_run = false;
        if(argc == 5) {
           prev_run_file = argv[4];
           do_edge_counts = true;
        }   
        else
           return usage();
    }
    else
        return usage();

    the_app_pid = pid;

    node_map_t nodes;
    edge_set_t edges;
    vector< pair<kptr_t, kptr_t> > indirect_calls;
    
    // Attach to the kernel on the target host
    if ((rv = kmgr.attach(host, port)) < 0) {
        cerr << "attach error\n";
        return rv;
    }

    // Turn off testing mode, just in case it was turned on by a 
    // previous kerninstd client
    kmgr.setTestingFlag(false);
    //kmgr.setTestingFlag(true);

    // Create caches of entry counters, both local and kernel
    createCtrCache();

    if (! start_run) // use previous results as start
       initializeFromPrevRun(prev_run_file, nodes, edges, 
                             indirect_calls, do_edge_counts);

    else // start at syscall entry point
       createRootNode(start_mod, start_func, nodes, indirect_calls);

    char *proc_pid_file = (char*)malloc(20);
    sprintf(proc_pid_file, "/proc/%d", pid);

    cg_file_name = (char*)malloc(20);
    sprintf(cg_file_name, "%d.dt", pid);

    // Loop until child finishes
    bool child_terminated = false;
    struct stat stat_buf;
    if( -1 == stat(proc_pid_file, &stat_buf) ) {
        printf("pid not valid, possibly terminated already\n");
        child_terminated = true;
    }
    else {
       
        printf("Starting call graph profiling for pid %d\n", pid);

        while(!child_terminated) {
            if(kmgr.handleEvents() < 0) {
                cerr << "handleEvents error\n";
                kmgr.detach();
                return -1;
            }
       
            if (!do_edge_counts) {

                if((watch_start == 0) && (indirect_calls.size() > 0)) {
                    //printf("PARENT: watch begin1\n");
                    watchIndirects(indirect_calls, watch_start);
                }
       
                // doProfile only appends to vectors
                //printf("PARENT: calling doProfile()\n");
                doProfile(&nodes, &edges, indirect_calls);
       
                if(watch_start) {
                    time_t curr = time(NULL);
                    if(curr - watch_start > callee_watching_delay) {
                        //printf("PARENT: watch end1\n");
                        unwatchIndirects(indirect_calls, &nodes, &edges);
                        //printf("PARENT: watch begin2\n");
                        watchIndirects(indirect_calls, watch_start);
                    }
                }
            }
       
            //printf("PARENT: checking for child termination\n");
            if( -1 == stat(proc_pid_file, &stat_buf) ) {
                //printf("PARENT: saw child terminate\n");
                child_terminated = true;
            }
            else {
               if(do_edge_counts)
                   sleep(30);
               else
                   sleep(profile_delay);
            }
        }
    
        if(do_edge_counts) {
            edge_set_t::const_iterator eiter = edges.begin();
            for (; eiter != edges.end(); eiter++) {
                if (kmgr.removeSnippet(eiter->inst_id) < 0) {
                    fprintf(stderr, "error uininstrumenting edge %d->%d\n",
                            eiter->site_addr, eiter->callee_addr);
                }
            }
        }   
        else {
            unwatchIndirects(indirect_calls, &nodes, &edges);
            node_map_t::iterator iter = nodes.begin();
            for (; iter != nodes.end(); iter++) {
                node_t *node = iter->second;
                if (node->unInstrument() < 0) {
                    fprintf(stderr, "error uninstrumenting node %d\n", 
                            node->getId());
                }
            }
        }


        // Print out results
        printf("Reading final counters\n");
        readCtrs();
        dumpCtrCache();

        if (dumpCallGraph(nodes, edges, do_edge_counts) < 0) {
            error_exit(-1);
        }
    }

    free(cg_file_name);
    
    freeCtrCache();
    
    // Detach from the kernel
    return kmgr.detach();
}
